package jp.agentec.sinaburocast.common.proc;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.List;
import java.util.Map;

import jp.agentec.sinaburocast.common.util.PropertyUtil;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * プロセス実行スレッドクラス
 *
 * @author tsukada
 *
 */
class ProcessExecutor extends Thread {
	private final Log logger= LogFactory.getLog(getClass());

	private StringBuffer stdOut;
	private StringBuffer stdErr;
	private ReaderWorker readerWorker;

	private int retVal = 0;

	private boolean completed = false; // timeoutせずにプロセスが終了したこと
	private boolean finished = false; // スレットの実行が終了したこと
	private Exception exception;
	private Process process;
	private List<String> command;

	@SuppressWarnings("unchecked")
	private Map env;

	private StreamPumper spStdout = null;
	private StreamPumper spStderr = null;
	private Thread readerWorkerThread = null;

	private Log processLogger;

	@SuppressWarnings("unchecked")
	ProcessExecutor(StringBuffer stdOut, StringBuffer stdErr, ReaderWorker readerWorker, Log processLogger, List<String> command, Map env) {
		super();
		this.stdOut = stdOut;
		this.stdErr = stdErr;
		this.readerWorker = readerWorker;
		this.processLogger = processLogger;
		this.command = command;
		this.env = env;
	}

	/**
	 * プロセス停止
	 */
	void destroyProcess() {
		if (process != null) {
			process.destroy();
		}
		if (spStdout != null) {
			spStdout.finish();
			spStdout.interrupt();
		}
		if (spStderr != null) {
			spStderr.finish();
			spStderr.interrupt();
		}
		if (readerWorkerThread != null) {
//			readerWorkerThread.finish();
			readerWorkerThread.interrupt();
		}
	}

	@Override
	public void run() {
		BufferedReader brInput = null;
		BufferedReader brError = null;
		try {
			ProcessBuilder builder = new ProcessBuilder(command);
			setEnv(env, builder);

			process = builder.start();

			brInput = new BufferedReader(new InputStreamReader(process.getInputStream(), PropertyUtil.getProperty("PROCESS_OUTPUT_CHARSET")));
			brError = new BufferedReader(new InputStreamReader(process.getErrorStream(), PropertyUtil.getProperty("PROCESS_OUTPUT_CHARSET")));

			if (readerWorker != null) {
				readerWorker.setBufferedReader(brInput);
				readerWorker.setCommandStr(command.toString());
				readerWorkerThread = new Thread(readerWorker);
				readerWorkerThread.start();
			}
			else {
				spStdout = new StreamPumper("STDOUT", brInput, stdOut, processLogger, "debug");
				spStdout.start();
			}

			spStderr = new StreamPumper("STDERR", brError, stdErr, processLogger, "debug");
			spStderr.start();

			// プロセスの終了を待機
			process.waitFor();

			retVal = process.exitValue();

			if (readerWorkerThread != null) {
				readerWorkerThread.join();
			}
			else if (spStdout != null) {
				spStdout.join();
			}
			spStderr.join();

//			process.destroy();
			logger.info("process exited with value : " + retVal);

			completed = true;

		} catch (InterruptedException e) {
			logger.error("Process error", e);
			exception = e;
		} catch (IOException e) {
			logger.error("Process error", e);
			exception = e;
		} finally {
			finished = true;
			try {
				if (brInput != null) {
					brInput.close();
				}
			} catch (final IOException e) {
				logger.error("buffered reader close error", e);
			}
			try {
				if (processLogger == null &&brError != null) {
					brError.close();
				}
			} catch (final IOException e) {
				logger.error("buffered error reader close error", e);
			}
		}
	}

	/**
	 * タイムアウトせずにプロセス実行が完了したかを返す。
	 *
	 * @return
	 */
	public boolean isCompleted() {
		return completed;
	}

	/**
	 * 発生した例外を返す。
	 * @return
	 */
	public Exception getException() {
		return exception;
	}

	/**
	 * 終了フラグを返す
	 * @return
	 */
	public boolean isFinished() {
		return finished;
	}

	/**
	 * プロセス実行結果を返す。
	 *
	 * @return
	 */
	public int getRetVal() {
		return retVal;
	}


	@SuppressWarnings("unchecked")
	private static void setEnv(Map env, ProcessBuilder builder) {
		if (env != null && !env.isEmpty()) {
			Map builderEnv = builder.environment();
			for(Object key : env.keySet().toArray()) {
				builderEnv.put(key, env.get(key));
			}
		}
	}


}
